from typing import Generator, Union, overload

from .basetypes import MODE_REGISTRY, Modes, ModesType
from .group_commands import GroupCommands

from ..command import Command


class GroupModes(Modes):
    def __init__(self):
        self.modes = MODE_REGISTRY
    
    def __iter__(self) -> Generator[Command, None, None]:
        for mode in self.modes.values():
            for command in mode:
                yield command

    @overload
    def __getitem__(self, key: str) -> Command: ...

    @overload
    def __getitem__(self, key: int) -> ModesType: ...

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, str):
            key = key.upper()
            item = getattr(self, key, None)
            if not isinstance(item, Command):
                raise KeyError(f"Command '{key}' not found")
            return item
        elif isinstance(key, int):
            mode = self.modes.get(key)
            if not isinstance(mode, GroupCommands):
                raise KeyError(f"Mode '{key}' not found")
            return mode

        raise TypeError(f"Invalid key type: {type(key)}. Expected str or int.")


from logging import getLogger
from typing import List, Tuple

from ..errors import ResponseBaseError
from ..mode import Mode
from ..protocol import Protocol
from ..response import ResponseBase, Response
from ..utils import bytes_to_string, filter_bytes, is_bytes_hexadecimal, split_by_byte

from .protocol_base import ProtocolBase


_log = getLogger(__name__)


class ProtocolCAN(ProtocolBase):
    """Supported Protocols:
    - [0x06] ISO 15765-4 CAN (11 bit ID, 500 Kbaud)
    - [0x07] ISO 15765-4 CAN (29 bit ID, 500 Kbaud)
    - [0x08] ISO 15765-4 CAN (11 bit ID, 250 Kbaud)
    - [0x09] ISO 15765-4 CAN (29 bit ID, 250 Kbaud)
    - [0x0A] SAE J1939 CAN (29 bit ID, 250 Kbaud)
    - [0x0B] USER1 CAN (11 bit ID, 125 Kbaud)
    - [0x0C] USER2 CAN (11 bit ID, 50 Kbaud)
    """
    def parse_response(self, response_base: ResponseBase) -> Response:
        context = response_base.context
        command = context.command
        if command.mode == Mode.AT: # AT Commands
            status = None
            if len(response_base.messages[:-1]) == 1:
                status = bytes_to_string(response_base.messages[0])

            return Response(**vars(response_base), value=status)
        else: # OBD Commands
            value = None
            parsed_data: List[Tuple[bytes, ...]] = list()
            for raw_line in response_base.messages[:-1]: # Skip the last line (prompt character)
                line = filter_bytes(raw_line, b' ')

                if not is_bytes_hexadecimal(line):
                    is_error = ResponseBaseError.detect(raw_line)
                    if not is_error:
                        continue
                    _log.error(is_error.message)
                    raise is_error

                attr = self.get_protocol_attributes(context.protocol)
                if not "header_length" in attr:
                    raise AttributeError(f"Missing required attribute 'header_length' in protocol attributes for protocol {context.protocol}")

                components = split_by_byte(line)

                if attr["header_length"] == 11: # Normalize to 29 bits (32 with hex)
                    components = (b"00",) * 2 + components

                minimal_length = 7 # Less means no data
                if len(components) < minimal_length:
                    _log.warning(f"Invalid line: too few components (expected at least {minimal_length}, got {len(components)})")
                    continue

                # header_end = 4 # unused
                length_idx = 4
                bytes_offset = 2
                response_idx = 5

                # header = b''.join(components[:header_end]) # unused
                length = int(components[length_idx], 16) - bytes_offset
                if length == 0:
                    continue
                response_code = int(components[response_idx], 16)
                data = components[-length:]

                if command.n_bytes and length != command.n_bytes:
                    _log.warning(f"Expected {command.n_bytes} bytes, but received {length} bytes for command {command}")

                if command.mode == Mode.REQUEST and not 0x40 + command.mode.value == response_code:
                    _log.warning(f"Unexpected response code 0x{response_code:02X} for command {command} (expected response code 0x{0x40 + command.mode.value:02X})")

                parsed_data.append(data)
            if command.formula:
                try:
                    value = command.formula(parsed_data)
                except Exception as e:
                    _log.error(f"Unexpected error during formula execution: {e}", exc_info=True)
                    value = None

            return Response(**vars(response_base), parsed_data=parsed_data, value=value)


ProtocolCAN.register({
    Protocol.ISO_15765_4_CAN:   {"header_length": 11},
    Protocol.ISO_15765_4_CAN_B: {"header_length": 29},
    Protocol.ISO_15765_4_CAN_C: {"header_length": 11},
    Protocol.ISO_15765_4_CAN_D: {"header_length" :29},
    Protocol.SAE_J1939_CAN:     {"header_length": 29},
    Protocol.USER1_CAN:         {"header_length": 11}, # 11 bits by default, can be extended
    Protocol.USER2_CAN:         {"header_length": 11}, # 11 bits by default, can be extended
})


from enum import Enum


class Protocol(Enum):
    UNKNOWN = -1
    """Unknown protocol"""

    AUTO = 0x00
    """Automatically determine the protocol"""
    SAE_J1850_PWM = 0x01
    """SAE J1850 PWM (41.6 kbaud)"""
    SAE_J1850_VPW = 0x02
    """SAE J1850 VPW (10.4 kbaud)"""
    ISO_9141_2 = 0x03
    """ISO 9141-2 (5 baud init, 10.4 kbaud)"""
    ISO_14230_4_KWP = 0x04
    """ISO 14230-4 KWP (5 baud init, 10.4 kbaud)"""
    ISO_14230_4_KWP_FAST = 0x05
    """ISO 14230-4 KWP (fast init, 10.4 kbaud)"""
    ISO_15765_4_CAN = 0x06
    """ISO 15765-4 CAN (11 bit ID, 500 kbaud)"""
    ISO_15765_4_CAN_B = 0x07
    """ISO 15765-4 CAN (29 bit ID, 500 kbaud)"""
    ISO_15765_4_CAN_C = 0x08
    """ISO 15765-4 CAN (11 bit ID, 250 kbaud)"""
    ISO_15765_4_CAN_D = 0x09
    """ISO 15765-4 CAN (29 bit ID, 250 kbaud)"""
    SAE_J1939_CAN = 0x0A
    """SAE J1939 CAN (29 bit ID, 250* kbaud), default settings (user adjustable)"""
    USER1_CAN = 0x0B
    """USER1 CAN (11* bit ID, 125* kbaud), default settings (user adjustable)"""
    USER2_CAN = 0x0C
    """USER2 CAN (11* bit ID, 50* kbaud), default settings (user adjustable)"""


from .protocol_can import ProtocolCAN
from .protocol_j1850 import ProtocolJ1850
from .protocol_kwp2000 import ProtocolKWP2000

__all__ = [
    "ProtocolCAN",
    "ProtocolJ1850",
    "ProtocolKWP2000",
]


import pytest

from obdii.modes import Mode01, Mode02, Mode03, Mode04, Mode09
from obdii.modes.group_modes import GroupModes


@pytest.mark.parametrize(
    "key, expected",
    [
        (1, Mode01()),
        (2, Mode02()),
        (3, Mode03()),
        (4, Mode04()),

        (9, Mode09()),

        (0, KeyError),
    ]
)
def test_commands_getitem_int(key, expected):
    commands = GroupModes()

    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            commands[key]
    else:
        result = commands[key]

        assert result == expected


@pytest.mark.parametrize(
    "key, expected",
    [
        ("SUPPORTED_PIDS_A", Mode01().SUPPORTED_PIDS_A),

        ('', KeyError),
    ]
)
def test_commands_getitem_str(key, expected):
    commands = GroupModes()

    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            commands[key]
    else:
        result = commands[key]

        assert result == expected


@pytest.mark.parametrize(
    "key_1, key_2, expected",
    [
        (1, 0, Mode01().SUPPORTED_PIDS_A),

        (0, 0, KeyError),
    ]
)
def test_commands_getitem_int_int(key_1, key_2, expected):
    commands = GroupModes()

    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            commands[key_1][key_2]
    else:
        result = commands[key_1][key_2]

        assert result == expected


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Type

from ..protocol import Protocol
from ..response import ResponseBase, Response


class ProtocolBase(ABC):
    _registry: Dict[Protocol, Type[ProtocolBase]] = {}
    _protocol_attributes: Dict[Protocol, Dict] = {}

    def __init__(self) -> None: ...

    @abstractmethod
    def parse_response(self, response_base: ResponseBase) -> Response: ...

    @classmethod
    def register(cls, protocols: Dict[Protocol, Dict[str, Any]]) -> None:
        """Register a subclass with its supported protocols."""
        for protocol, attr in protocols.items():
            cls._registry[protocol] = cls
            cls._protocol_attributes[protocol] = attr

    @classmethod
    def get_handler(cls, protocol: Protocol) -> ProtocolBase:
        """Retrieve the appropriate protocol class or fallback to ProtocolUnknown."""
        return cls._registry.get(protocol, ProtocolUnknown)()

    @classmethod
    def get_protocol_attributes(cls, protocol: Protocol) -> Dict[str, Any]:
        return cls._protocol_attributes.get(protocol, {})


class ProtocolUnknown(ProtocolBase): 
    """Fallback protocol class for unknown or unsupported protocols.

    In such cases, basic serial communication might still be possible,
    but full message parsing could be limited.
    """
    def parse_response(self, response_base: ResponseBase) -> Response:
        raise NotImplementedError


import pytest

from obdii.errors import (
    ResponseBaseError, 
    InvalidCommandError, BufferFullError, BusBusyError, BusError, 
    CanError, InvalidDataError, InvalidLineError, DeviceInternalError, SignalFeedbackError, 
    MissingDataError, CanDataError, StoppedError, ProtocolConnectionError,
    InactivityWarning, LowPowerWarning, LowVoltageResetWarning,
)


@pytest.mark.parametrize(
    "response, expected_error",
    [
        (b'?', InvalidCommandError),
        (b"BUFFER FULL", BufferFullError),
        (b"BUS BUSY", BusBusyError),
        (b"BUS ERROR", BusError),
        (b"CAN ERROR", CanError),
        (b"DATA ERROR", InvalidDataError),
        (b"SOME DATA ERROR", InvalidDataError),
        (b"<DATA ERROR", InvalidLineError),
        (b"SOME <DATA ERROR", InvalidLineError),
        (b"ERR01", DeviceInternalError),
        (b"ERR99", DeviceInternalError),
        (b"FB ERROR", SignalFeedbackError),
        (b"NO DATA", MissingDataError),
        (b"<RX ERROR", CanDataError),
        (b"STOPPED", StoppedError),
        (b"UNABLE TO CONNECT", ProtocolConnectionError),

        (b"ACT ALERT", InactivityWarning),
        (b"LP ALERT", LowPowerWarning),
        (b"LV RESET", LowVoltageResetWarning),

        (b"NO ERROR HERE", None),
    ]
)
def test_error_detection(response, expected_error):
    error = ResponseBaseError.detect(response)
    if expected_error is None:
        assert error is None, f"Expected no error but got {error}"
    else:
        assert isinstance(error, expected_error), f"Expected {expected_error.__name__}, but got {type(error).__name__}"


from ..protocol import Protocol
from ..response import ResponseBase, Response

from .protocol_base import ProtocolBase


class ProtocolKWP2000(ProtocolBase):
    """Supported Protocols:
    - [0x03] ISO 9141-2 (5 baud init, 10.4 Kbaud)
    - [0x04] ISO 14230-4 KWP (5 baud init, 10.4 Kbaud)
    - [0x05] ISO 14230-4 KWP (fast init, 10.4 Kbaud)
    """
    def parse_response(self, response_base: ResponseBase) -> Response:
        raise NotImplementedError


ProtocolKWP2000.register({
    Protocol.ISO_9141_2:           {},
    Protocol.ISO_14230_4_KWP:      {},
    Protocol.ISO_14230_4_KWP_FAST: {},
})


from typing import Generator, Union

from ..command import Command


class GroupCommands():
    def __getitem__(self, key) -> Command:
        if isinstance(key, int):
            for attr_name in dir(self):
                attr = getattr(self, attr_name)
                if isinstance(attr, Command) and attr.pid == key:
                    return attr
        raise KeyError(f"No command found with PID {key}")

    def __repr__(self) -> str:
        return f"<Mode Commands: {len(self)}>"

    def __len__(self) -> int:
        return sum(1 for attr_name in dir(self) if isinstance(getattr(self, attr_name), Command))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, GroupCommands):
            return False
        
        return vars(self) == vars(value)

    def __iter__(self) -> Generator[Command, None, None]:
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if isinstance(attr, Command):
                yield attr
    
    def has_command(self, command: Union[Command, str]) -> bool:
        if isinstance(command, Command):
            command = command.name

        command = command.upper()
        return isinstance(getattr(self, command, None), Command)


from socket import AF_INET, SOCK_STREAM, error as s_error, socket
from typing import Optional, Union

from .transport_base import TransportBase

from ..basetypes import MISSING
from ..utils import override_class_attributes


class TransportWifi(TransportBase):
    def __init__(self, **kwargs) -> None:
        self.overridable_attributes = {
            "address": MISSING,
            "port": MISSING,
            "timeout": 5.0,
        }

        self.address: str
        self.port: Union[str, int]
        self.timeout: float

        self.socket_conn: Optional[socket] = None

        override_class_attributes(self, self.overridable_attributes, **kwargs)

        if self.address is MISSING or self.port is MISSING:
            raise ValueError("Both address and port must be specified for TransportWifi.")


    def __repr__(self) -> str:
        return f"<TransportWifi {self.address}:{self.port}>"


    def is_connected(self) -> bool:
        if self.socket_conn is None:
            return False
        try:
            self.socket_conn.getpeername()
            return True
        except s_error:
            return False


    def connect(self, **kwargs) -> None:
        override_class_attributes(self, self.overridable_attributes, True, **kwargs)

        self.socket_conn = socket(AF_INET, SOCK_STREAM)
        self.socket_conn.settimeout(self.timeout)
        self.socket_conn.connect((self.address, int(self.port)))

    def close(self) -> None:
        if self.socket_conn:
            self.socket_conn.close()
        self.socket_conn = None


    def write_bytes(self, query: bytes) -> None:
        if not self.socket_conn:
            raise RuntimeError("Socket is not connected.")
        self.socket_conn.sendall(query)

    def read_bytes(self, expected_seq: bytes = b'>', size: int = MISSING) -> bytes:
        if not self.socket_conn:
            raise RuntimeError("Socket is not connected.")

        lenterm = len(expected_seq)
        buffer = bytearray()
        while True:
            chunk = self.socket_conn.recv(1)
            if not chunk:
                raise RuntimeError("Socket connection closed.")
            
            buffer += chunk

            if buffer[-lenterm:] == expected_seq:
                break

            if size is not MISSING and len(buffer) >= size:
                break

        return bytes(buffer)



from copy import deepcopy
from re import findall
from typing import Any, Callable, Dict, Optional, Union

from .basetypes import MISSING, OneOrMany, Real
from .mode import Mode


class Command():
    def __init__(self, 
            mode: Mode,
            pid: Union[int, str],
            n_bytes: int,
            name: str,
            description: str = MISSING,
            min_values: Optional[OneOrMany[Real]] = MISSING,
            max_values: Optional[OneOrMany[Real]] = MISSING,
            units: Optional[OneOrMany[str]] = MISSING,
            formula: Optional[Callable] = MISSING,
            command_args: Optional[Dict[str, Any]] = MISSING,
        ) -> None:
        """
        Initializes a Command instance with the given parameters.

        Parameters
        ----------
        mode: :class:`Mode`
            Command mode to be used.
        pid: Union[:class:`int`, :class:`str`]
            Command PID (Parameter Identifier) to be used.
        n_bytes: :class:`int`
            The number of bytes expected in the response.
        name: :class:`str`
            The name of the command.
        description: :class:`str`
            A description of the command.
        min_values: Optional[Union[:class:`int`, :class:`float`, List[Union[:class:`int`, :class:`float`]]]]
            The minimum valid values for the command's parameters.
        max_values: Optional[Union[:class:`int`, :class:`float`, List[Union[:class:`int`, :class:`float`]]]]
            The maximum valid values for the command's parameters.
        units: Optional[Union[:class:`str`, List[:class:`str`]]]
            The units for the command's response.
        formula: Optional[:class:`Callable`]
            A formula for transforming the response value.
        command_args: Optional[Dict[:class:`str`, Any]]
            A dictionary containing the argument names and their expected types for formatting the command.
        """
        self.mode = mode
        self.pid = pid
        self.n_bytes = n_bytes
        self.name = name
        self.description = description
        self.min_values = min_values
        self.max_values = max_values
        self.units = units
        self.formula = formula

        self.command_args = command_args or {}
        self.is_formatted = False

    def __call__(self, *args: Any, checks: bool = True) -> "Command":
        """
        Formats the command with the provided arguments and checks for validity.

        Parameters
        ----------
        *args: :class:`Any`
            The values for the arguments of the command, in the order they are defined in `command_args`.
        checks: :class:`bool`
            If `True`, performs validation checks on the arguments. Defaults to `True`.

        Returns
        -------
        :class:`Command`
            A new `Command` object with the formatted PID.

        Raises
        ------
        ValueError
            If the number of arguments does not match the expected number or if the placeholders are mismatched.
        TypeError
            If the argument type does not match the expected type.
        """
        if not self.command_args:
            raise ValueError(f"Command '{self}' should not be parametrized, as no arguments have been described.")

        if len(args) != len(self.command_args):
            raise ValueError(f"Expected {len(self.command_args)} arguments, got {len(args)}.")

        placeholders = set(findall(r"{(\w+)}", str(self.pid)))
        expected_placeholders = set(self.command_args.keys())

        if placeholders != expected_placeholders:
            missing = expected_placeholders - placeholders
            extra = placeholders - expected_placeholders
            raise ValueError(f"PID format mismatch. Missing placeholders: {missing}. Extra placeholders: {extra}.")

        combined_args = {}
        for (arg_name, arg_type), value in zip(self.command_args.items(), args):
            if checks:
                if not isinstance(value, arg_type):
                    raise TypeError(f"Expected argument '{arg_name}' to be of type {arg_type.__name__}, got {type(value).__name__}.")

                expected_len = len(arg_name)

                if isinstance(value, int):
                    if value < 0:
                        raise ValueError(f"Argument '{arg_name}' cannot be negative.")
                    value = f"{value:0{expected_len}X}"
                    if len(value) != expected_len:
                        raise ValueError(f"Argument '{arg_name}' must be {expected_len} characters long after formatting, got {len(value)}.")

                elif isinstance(value, str):
                    if len(value) != expected_len:
                        raise ValueError(f"Argument '{arg_name}' must be {expected_len} characters long, got {len(value)}.")

                else:
                    raise TypeError(f"Argument '{arg_name}' must be of type int or str, got {type(value).__name__}.")

            combined_args[arg_name] = value

        fmt_command = deepcopy(self)
        fmt_command.is_formatted = True
        fmt_command.pid = str(self.pid).format(**combined_args)

        return fmt_command

    def __repr__(self) -> str:
        return f"<Command {self.mode} {self.pid if isinstance(self.pid, str) else f'{self.pid:02X}'} {self.name or 'Unnamed'} [{', '.join(self.command_args.keys())}]>"
    
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Command):
            return False

        return vars(self) == vars(value)
    
    def __hash__(self) -> int:
        return hash((self.mode, self.pid, self.name))

    def build(self, early_return: bool = False) -> bytes:
        """
        Builds the query to be sent to the ELM327 device as a byte string.
        (The ELM327 is case-insensitive, ignores spaces and all control characters.)

        Parameters
        ----------
        early_return: :class:`bool`
            If set to `True`, appends a hex digit representing the expected number of responses in the query. 
            Defaults to `False`.

        Returns
        -------
        :class:`bytes`
            The formatted query as a byte string, ready to be sent to the ELM327 device.

        Raises
        ------
        ValueError
            If arguments have not been set or are incorrectly formatted.
        """
        if self.command_args and not self.is_formatted:
            raise ValueError(f"Command has unset arguments for '{self.pid}': {self.command_args}")

        mode = self.mode.value
        pid = self.pid
        return_digit = ''
        if early_return and self.n_bytes and self.mode != Mode.AT:
            data_bytes = 7

            n_lines = (self.n_bytes + (data_bytes - 1)) // data_bytes
            if 0 < n_lines < 16:
                return_digit = f" {n_lines:X}"

        if isinstance(mode, int):
            mode = f"{mode:02X}"
        if isinstance(pid, int):
            pid = f"{pid:02X}"

        query = f"{mode} {pid}{return_digit}\r"

        return query.encode()


from typing import Any, Dict, Iterable, List, Literal, Tuple, TypeVar, Union


T = TypeVar('T')

class SingletonMeta(type):
    _instances: Dict[type, object] = {}

    def __call__(cls, *args, **kwargs) -> object:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

class _MissingSentinel(metaclass=SingletonMeta):
    __slots__ = ()

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, _MissingSentinel)

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> Literal["..."]:
        return "..."

MISSING: Any = _MissingSentinel()

OneOrMany = Union[T, Iterable[T]]

Real = Union[int, float]

BytesRows = List[Tuple[bytes, ...]]


from copy import deepcopy

from .group_commands import GroupCommands
from .mode_01 import Mode01

from ..command import Command
from ..mode import Mode


# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_02_-_Show_freeze_frame_data

class Mode02(GroupCommands):
    """Freeze frame Commands"""

    @classmethod
    def _populate_commands(cls):
        for field_name, command in vars(Mode01).items():
            if isinstance(command, Command):
                field = f"DTC_{field_name}"
                dtc_command = deepcopy(command)
                dtc_command.mode = Mode.FREEZE_FRAME
                dtc_command.name = field
                setattr(Mode02, field, dtc_command)

Mode02._populate_commands()


from functools import partial

from .group_commands import GroupCommands

from ..command import Command
from ..mode import Mode


M = Mode.CLEAR_DTC
C = partial(Command, M)

# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_04_-_Clear_Diagnostic_Trouble_Codes_and_stored_values

class Mode04(GroupCommands):
    """Clear Diagnostic Trouble Codes Command"""
    CLEAR_DTC = C('', 0x00, "CLEAR_DTC", "Clear trouble codes / Malfunction indicator lamp (MIL) / Check engine light", None, None, None)


from ast import (
    Add, Sub, Mult, Div, FloorDiv, Mod, Pow, BitXor, UAdd, USub, parse, NodeVisitor, Expr, BinOp, UnaryOp, Name, Constant
)
from typing import List, Any, NoReturn
from operator import add, sub, mul, truediv, floordiv, mod, pow, xor

from ..basetypes import BytesRows


class SafeEvaluator(NodeVisitor):
    """A safe evaluator that only allows basic math operations."""
    
    operators = {
        Add: add, Sub: sub,
        Mult: mul, Div: truediv,
        FloorDiv: floordiv, Mod: mod,
        Pow: pow, BitXor: xor
    }

    def __init__(self, variables):
        self.variables = variables

    def visit_BinOp(self, node: BinOp) -> Any:
        return self.operators[type(node.op)](self.visit(node.left), self.visit(node.right))

    def visit_UnaryOp(self, node: UnaryOp) -> Any:
        operand = self.visit(node.operand)
        if isinstance(node.op, UAdd):
            return +operand
        elif isinstance(node.op, USub):
            return -operand
        raise ValueError("Unsupported operation")

    def visit_Name(self, node: Name) -> Any:
        if node.id in self.variables:
            return self.variables[node.id]
        raise ValueError(f"Unknown variable: {node.id}")

    def visit_Constant(self, node: Constant) -> Any:
        return node.value

    def visit_Expr(self, node: Expr) -> Any:
        return self.visit(node.value)

    def generic_visit(self, node) -> NoReturn:
        raise ValueError("Unsupported operation in formula")


class Formula():
    def __init__(self, expression: str):
        """Initialize with a formula string and extract variable names dynamically."""
        self.expression = expression.upper()

        self.parsed_expr = parse(self.expression, mode="eval")

    def __call__(self, parsed_data: BytesRows) -> Any:
        """Evaluate the formula."""
        if not parsed_data or not parsed_data[0]:
            raise ValueError("Invalid parsed_data: must contain at least one non-empty list.")

        first_item = parsed_data[0]
        
        values = {
            chr(ord('A') + i): int(first_item[i], 16) 
            for i in range(len(first_item))
        }

        evaluator = SafeEvaluator(values)
        return evaluator.visit(self.parsed_expr.body)


class MultiFormula():
    def __init__(self, *expressions: str) -> None:
        """Initialize multiple Formula objects from a list of expressions."""
        self.formulas = [Formula(expr) for expr in expressions]

    def __call__(self, parsed_data: BytesRows) -> List[Any]:
        """Evaluate all stored formulas on the given parsed_data."""
        return [formula(parsed_data) for formula in self.formulas]


from csv import DictWriter
from datetime import datetime
from logging import FileHandler, DEBUG, getLogger
from time import time, sleep

from obdii import Connection, Protocol, at_commands, commands
from obdii.errors import InvalidCommandError

description = """ Connects to an OBD device through serial port, using custom and advanced startup.
Fetch and log a batch of specified commands every second in a CSV."""


# File logging
file_handler = FileHandler("obdii.log")
obdii_logger = getLogger("obdii")

port = "COM5"  # Replace with the actual serial port

# Commands to log every second
batch_cmd = [
    commands.ENGINE_SPEED,
    commands.VEHICLE_SPEED,
    commands.ENGINE_OIL_TEMP,
    commands.ENGINE_COOLANT_TEMP,
]
max_runtime = 15
csv_file = "obd_data.csv"

with Connection(
        port,
        protocol = Protocol.ISO_15765_4_CAN,
        auto_connect = False,
        early_return = True,

        log_handler = file_handler,
        log_level = DEBUG,
    ) as conn:
    # Override the default init_sequence
    conn.init_sequence = [
        at_commands.RESET,
        at_commands.ECHO_OFF,
        at_commands.HEADERS_ON,
        at_commands.MEMORY_OFF,
        at_commands.SET_PROTOCOL(0), # Usage of command argument(s)
        conn._auto_protocol, # Detect and set required protocol if available
    ]

    # Advanced use of the connect method
    conn.connect(
        baudrate = 38400,
        timeout = 10,
    )

    # Write in a csv cmds values within the max_runtime
    with open(csv_file, 'w', newline='') as file:
        cmd_keys = [cmd.name for cmd in batch_cmd]
        fieldnames = ["timestamp"] + cmd_keys
        writer = DictWriter(file, fieldnames=fieldnames)
        writer.writeheader()

        start_time = time()
        while conn.is_connected() and (time() - start_time) < max_runtime:
            log_entry = dict.fromkeys(cmd_keys, None)
            log_entry["timestamp"] = datetime.now().isoformat()

            for cmd in batch_cmd:
                try:
                    response = conn.query(cmd)
                    log_entry[cmd.name] = response.value
                except InvalidCommandError:
                    obdii_logger.warning(f"{cmd.name} seems not to be supported.")

            writer.writerow(log_entry)
            sleep(1)


from dataclasses import dataclass, field
from time import time
from typing import Any, List, Optional

from .basetypes import BytesRows, OneOrMany, Real
from .command import Command
from .protocol import Protocol


@dataclass
class Context():
    command: Command
    protocol: Protocol
    timestamp: float = field(default_factory=time)


@dataclass
class ResponseBase():
    context: Context
    raw: bytes
    messages: List[bytes]
    timestamp: float = field(default_factory=time)


@dataclass
class Response(ResponseBase):
    parsed_data: Optional[BytesRows] = None

    value: Optional[Any] = None

    @property
    def min_values(self) -> Optional[OneOrMany[Real]]:
        return self.context.command.min_values
    
    @property
    def max_values(self) -> Optional[OneOrMany[Real]]:
        return self.context.command.max_values
    
    @property
    def units(self) -> Optional[OneOrMany[str]]:
        return self.context.command.units



from obdii import Connection, at_commands, commands

description = """Simple interaction with an OBD device.
Demonstrates different approaches to query commands."""


# Establish a connection to the serial port (e.g., COM5)
conn = Connection("COM5")


# Query the ELM327 version using an AT command
version = conn.query(at_commands.VERSION_ID)

print(f"Version: {version.value}")


# Different ways to request data
# 1. Using a command directly
response = conn.query(commands.ENGINE_SPEED)
print(f"1. Base command: {response.value} {response.units} (Engine speed)")

# 2. Using the command name
response = conn.query(commands["ENGINE_SPEED"])
print(f"2. Command name: {response.value} {response.units} (Engine speed)")

# 3. Using the Mode and PID
response = conn.query(commands[1][0x0C])
print(f"3. Mode and PID: {response.value} {response.units} (Engine speed)")


# Close the connection
conn.close()


from logging import Handler, Formatter, DEBUG, INFO, WARNING, ERROR, CRITICAL, StreamHandler, getLogger
from os import environ
from re import escape, fullmatch, sub
from sys import platform
from typing import Any, Dict, Tuple

from .basetypes import MISSING
from .response import ResponseBase


def split_by_byte(raw: bytes) -> Tuple[bytes, ...]:
    if len(raw) % 2 != 0:
        raw = b'0' + raw

    return tuple(raw[i:i+2] for i in range(0, len(raw), 2))


def is_bytes_hexadecimal(raw: bytes) -> bool:
    return bool(fullmatch(b"[0-9A-Fa-f]+", raw))


def filter_bytes(raw: bytes, *filter_bytes: bytes) -> bytes:
    pattern = b'|'.join(escape(filter_byte) for filter_byte in filter_bytes)

    return sub(pattern, b'', raw)


def bytes_to_string(raw: bytes) -> str:
    return raw.decode(errors="ignore").strip()


def debug_responsebase(response_base: ResponseBase) -> str:
    return '\n'.join(f"[{bytes_to_string(line)}]" for line in response_base.messages[:-1]) + '\n'


def override_class_attributes(cls: object, attributes: Dict[str, Any], pop: bool = False, **kwargs) -> None:
    """Override class attributes from kwargs."""
    for key, value in attributes.items():
        setattr(
            cls, 
            key, 
            kwargs.get(key, value) if not pop else kwargs.pop(key, getattr(cls, key, value))
                )


def setup_logging(
    handler: Handler = MISSING,
    formatter: Formatter = MISSING,
    level: int = MISSING,
    root: bool = True,
) -> None:
    """A helper function to setup logging."""
    if level is MISSING:
        level = INFO

    if handler is MISSING:
        handler = StreamHandler()

    if formatter is MISSING:
        if isinstance(handler, StreamHandler) and _stream_supports_colour(handler.stream):
            formatter = _ColorFormatter()
        else:
            dt_fmt = "%Y-%m-%d %H:%M:%S"
            formatter = Formatter("[{asctime}] [{levelname:<8}] {name}: {message}", dt_fmt, style='{')

    if root:
        logger = getLogger()
    else:
        library, _, _ = __name__.partition('.')
        logger = getLogger(library)

    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)

def _stream_supports_colour(stream: Any) -> bool:
    is_a_tty = hasattr(stream, "isatty") and stream.isatty()

    if "PYCHARM_HOSTED" in environ or environ.get("TERM_PROGRAM") == "vscode":
        return is_a_tty

    if platform != "win32":
        return is_a_tty

    return is_a_tty and "WT_SESSION" in environ

class _ColorFormatter(Formatter):
    LEVEL_COLORS = [
        (DEBUG, "\x1b[40;1m"),
        (INFO, "\x1b[34;1m"),
        (WARNING, "\x1b[33;1m"),
        (ERROR, "\x1b[31m"),
        (CRITICAL, "\x1b[41m"),
    ]

    FORMATS = {
        level: Formatter(
            f"\x1b[30;1m%(asctime)s\x1b[0m {colour}%(levelname)-8s\x1b[0m \x1b[35m%(name)s\x1b[0m %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
        for level, colour in LEVEL_COLORS
    }

    def format(self, record):
        formatter = self.FORMATS.get(record.levelno)
        if formatter is None:
            formatter = self.FORMATS[DEBUG]

        if record.exc_info:
            text = formatter.formatException(record.exc_info)
            record.exc_text = f"\x1b[31m{text}\x1b[0m"

        output = formatter.format(record)

        record.exc_text = None
        return output



from ..protocol import Protocol
from ..response import ResponseBase, Response

from .protocol_base import ProtocolBase


class ProtocolJ1850(ProtocolBase):
    """Supported Protocols:
    - [0x01] SAE J1850 PWM (41.6 Kbaud)
    - [0x02] SAE J1850 VPW (10.4 Kbaud)
    """
    def parse_response(self, response_base: ResponseBase) -> Response:
        raise NotImplementedError


ProtocolJ1850.register({
    Protocol.SAE_J1850_PWM: {},
    Protocol.SAE_J1850_VPW: {},
})


import pytest

from inspect import getsource
from ast import parse, walk, Assign, Name

from obdii.command import Command
from obdii.modes import ModeAT, Mode01, Mode02, Mode03, Mode04, Mode09

TEST_MODES = [
    ModeAT,
    Mode01,
    Mode02,
    Mode03,
    Mode04,

    Mode09,
]


@pytest.mark.parametrize(
    "mode",
    TEST_MODES,
)
def test_field_name_matches_command_name(mode):
    for field_name, field_value in vars(mode).items():
        if isinstance(field_value, Command):
            assert field_value.name == field_name, f"Field '{field_value.name}' does not match with the command name '{field_value.name}'."

@pytest.mark.parametrize(
    "mode",
    TEST_MODES,
)
def test_mins_maxs_units(mode):
    for command in vars(mode).values():
        if isinstance(command, Command):
            min_vals = command.min_values
            max_vals = command.max_values
            units = command.units

            if isinstance(max_vals, (list, tuple)) or isinstance(min_vals, (list, tuple)) or isinstance(units, (list, tuple)):
                assert isinstance(max_vals, (list, tuple)), f"Expected list/tuple for max_values, got {type(max_vals)}"
                assert isinstance(min_vals, (list, tuple)), f"Expected list/tuple for min_values, got {type(min_vals)}"
                assert isinstance(units, (list, tuple)), f"Expected list/tuple for units, got {type(units)}"

                # Ensure they have the same length
                assert len(max_vals) == len(min_vals) == len(units), f"Length mismatch: max={len(max_vals)}, min={len(min_vals)}, units={len(units)}"

@pytest.mark.parametrize(
    "mode",
    TEST_MODES,
)
def test_for_unique_fields(mode):
    source = getsource(mode)
    tree = parse(source)
    assignments = {}
    duplicates = []

    for node in walk(tree):
        if isinstance(node, Assign):
            for target in node.targets:
                if isinstance(target, Name):
                    var_name = target.id
                    if var_name in assignments:
                        duplicates.append(var_name)
                    assignments[var_name] = True
    
    assert not duplicates, f"Duplicate field(s) defined in {mode.__name__}: {', '.join(duplicates)}"



from typing import Dict, Union

from .mode_01 import Mode01
from .mode_02 import Mode02
from .mode_03 import Mode03
from .mode_04 import Mode04

from .mode_09 import Mode09


ModesType = Union[
    Mode01, 
    Mode02,
    Mode03,
    Mode04,

    Mode09,
    ]

class Modes(
    Mode01,
    Mode02,
    Mode03,
    Mode04,

    Mode09,
    ): ...

MODE_REGISTRY: Dict[int, ModesType] = {
    0x01: Mode01(),
    0x02: Mode02(),
    0x03: Mode03(),
    0x04: Mode04(),

    0x09: Mode09(),
}


from .group_modes import GroupModes

from .mode_at import ModeAT

from .mode_01 import Mode01
from .mode_02 import Mode02
from .mode_03 import Mode03
from .mode_04 import Mode04

from .mode_09 import Mode09


at_commands = ModeAT()
commands = GroupModes()

__all__ = [
    "at_commands",
    "commands",
    "ModeAT",

    "Mode01",
    "Mode02",
    "Mode03",
    "Mode04",

    "Mode09",
]


from logging import Formatter, Handler, getLogger, INFO
from re import IGNORECASE, search as research
from types import TracebackType
from typing import Callable, List, Optional, Tuple, Union

from .basetypes import MISSING
from .command import Command
from .modes import ModeAT
from .protocol import Protocol
from .protocols.protocol_base import ProtocolBase
from .response import Context, Response, ResponseBase
from .transports.transport_base import TransportBase
from .transports import TransportPort, TransportWifi
from .utils import bytes_to_string, debug_responsebase, filter_bytes, setup_logging


_log = getLogger(__name__)


class Connection():
    def __init__(self, 
                    transport: Union[str, Tuple[str, Union[str, int]], TransportBase],
                    protocol: Protocol = Protocol.AUTO,
                    auto_connect: bool = True,
                    smart_query: bool = False,
                    early_return: bool = False,
                    *,
                    log_handler: Handler = MISSING,
                    log_formatter: Formatter = MISSING,
                    log_level: int = INFO,
                    log_root: bool = False,
                    **kwargs,
                ) -> None:
        """
        Initialize connection settings and optionally auto-connect.

        Parameters
        ----------
        transport: Union[:class:`str`, Tuple[:class:`str`, Union[:class:`str`, :class:`int`]], :class:`obdii.transports.transport_base.TransportBase`]
            Can be represented as a string for serial ports (e.g., "COM5", "/dev/ttyUSB0", "/dev/rfcomm0"),
            or as a tuple for network transports (e.g., ("<hostname>", <port>)),
            or as an instance of a subclass of :class:`obdii.transports.transport_base.TransportBase`.
        protocol: :class:`Protocol`
            The protocol to use for communication.
        auto_connect: :class:`bool`
            If True, connect to the adapter immediately.
        smart_query: :class:`bool`
            If True, send repeat command when the same command is issued again.
        early_return: :class:`bool`
            If set to true, the ELM327 will return immediately after sending the specified number of responses specified in the command (n_bytes). Works only with ELM327 v1.3 and later.
        
        log_handler: :class:`logging.Handler`
            Custom log handler for the logger.
        log_formatter: :class:`logging.Formatter`
            Formatter to use with the given log handler.
        log_level: :class:`int`
            Logging level for the logger.
        log_root: :class:`bool`
            Whether to set up the root logger.
        
        **kwargs: :class:`dict`
            Additional keyword arguments to pass to the transport's connect method. Only used if `auto_connect` is True.
        """
        self.transport = self._resolve_transport(transport, **kwargs)

        self.protocol = protocol
        self.smart_query = smart_query
        self.early_return = early_return

        self.protocol_handler = ProtocolBase.get_handler(Protocol.UNKNOWN)
        self.supported_protocols: List[Protocol] = []
        self.last_command: Optional[Command] = None

        self.init_sequence: List[Union[Command, Callable[[], None]]] = [
            ModeAT.RESET,
            ModeAT.ECHO_OFF,
            ModeAT.HEADERS_ON,
            ModeAT.SPACES_ON,
            self._auto_protocol,
        ]
        self.init_completed = False

        self.protocol_preferences = [
            Protocol.ISO_15765_4_CAN,       # 0x06
            Protocol.ISO_15765_4_CAN_B,     # 0x07
            Protocol.ISO_15765_4_CAN_C,     # 0x08
            Protocol.ISO_15765_4_CAN_D,     # 0x09
            Protocol.SAE_J1850_PWM,         # 0x01
            Protocol.SAE_J1850_VPW,         # 0x02
            Protocol.ISO_9141_2,            # 0x03
            Protocol.ISO_14230_4_KWP,       # 0x04
            Protocol.ISO_14230_4_KWP_FAST,  # 0x05 
            Protocol.SAE_J1939_CAN,         # 0x0A
            Protocol.USER1_CAN,             # 0x0B
            Protocol.USER2_CAN,             # 0x0C
        ]

        if log_handler or log_formatter or log_level:
            setup_logging(
                log_handler,
                log_formatter,
                log_level,
                log_root
            )

        if auto_connect:
            self.connect(**kwargs)
    

    def _resolve_transport(self, transport: Union[str, Tuple[str, Union[str, int]], TransportBase], **kwargs) -> TransportBase:
        """Resolves a user-supplied transport input into a concrete TransportBase instance."""
        if isinstance(transport, str):
            return TransportPort(port=transport, **kwargs)
        elif isinstance(transport, tuple) \
            and len(transport) == 2 \
            and isinstance(transport[0], str) \
            and isinstance(transport[1], (str, int)):
            return TransportWifi(address=transport[0], port=transport[1], **kwargs)
        elif isinstance(transport, TransportBase):
            return transport
        else:
            raise TypeError(f"Invalid transport type: {type(transport)}.")


    def connect(self, **kwargs) -> None:
        """
        Establishes a connection to the device using the configured transport and runs the initialization sequence.

        Parameters
        ----------
        **kwargs
            Additional parameters forwarded to the transport's `connect()` method.
        """
        _log.info(f"Attempting to connect to {repr(self.transport)}.")
        try:
            self.transport.connect(**kwargs)
            self._initialize_connection()
            self.init_completed = True
            _log.info(f"Successfully connected to {repr(self.transport)}.")
        except Exception as e:
            self.transport.close()
            _log.error(f"Failed to connect to {repr(self.transport)}: {e}")
            raise ConnectionError(f"Failed to connect: {e}")

    def _initialize_connection(self) -> None:
        """Initializes the connection using the init sequence."""
        for command in self.init_sequence:
            if isinstance(command, Command):
                self.query(command)
            elif callable(command):
                command()
            else:
                _log.error(f"Invalid type in init_sequence: {type(command)}")
                raise TypeError(f"Invalid command type: {type(command)}")

    def is_connected(self) -> bool:
        """
        Checks if the transport connection is open.

        Returns
        -------
        :class:`bool`
            True if the connection is active.
        """
        return self.transport.is_connected()


    def _auto_protocol(self, protocol: Protocol = MISSING) -> None:
        """Sets the protocol for communication."""
        protocol = protocol or self.protocol
        unwanted_protocols = [Protocol.AUTO, Protocol.UNKNOWN]

        protocol_number = self._set_protocol_to(protocol)

        if Protocol(protocol_number) in unwanted_protocols:
            self.supported_protocols = self._get_supported_protocols()
            supported_protocols = self.supported_protocols

            if supported_protocols:
                priority_dict = {protocol: idx for idx, protocol in enumerate(self.protocol_preferences)}
                supported_protocols.sort(key=lambda p: priority_dict.get(p, len(self.protocol_preferences)))

                protocol_number = self._set_protocol_to(supported_protocols[0])
            else:
                protocol_number = -1

        self.protocol = Protocol(protocol_number)
        self.protocol_handler = ProtocolBase.get_handler(self.protocol)
        if protocol not in unwanted_protocols and protocol != self.protocol:
            _log.warning(f"Requested protocol {protocol.name} cannot be used.")
        _log.info(f"Protocol set to {self.protocol.name}.")

    def _set_protocol_to(self, protocol: Protocol) -> int:
        """Attempts to set the protocol to the specified value, return the protocol number if successful."""
        self.query(ModeAT.SET_PROTOCOL(protocol.value))
        response = self.query(ModeAT.DESC_PROTOCOL_N)

        line = bytes_to_string(filter_bytes(response.raw, b'\r', b'>'))
        protocol_number = self._parse_protocol_number(line)

        return protocol_number

    def _get_supported_protocols(self) -> List[Protocol]:
        """Attempts to find supported protocol(s)."""
        supported_protocols = []

        for protocol in Protocol:
            if protocol in [Protocol.UNKNOWN, Protocol.AUTO]:
                continue

            protocol_number = self._set_protocol_to(protocol)
            if protocol_number == protocol.value:
                supported_protocols.append(protocol)
        
        if not supported_protocols:
            _log.warning("No supported protocols detected.")
            supported_protocols = [Protocol.UNKNOWN]

        return supported_protocols

    def _parse_protocol_number(self, line: str) -> int:
        """Extracts and returns the protocol number from the response line."""
        match = research(r"([0-9A-F])$", line, IGNORECASE)
        if match:
            return int(match.group(1), 16)
        return -1


    def query(self, command: Command) -> Response:
        """
        Send a command and wait for the response.

        Parameters
        ----------
        command: :class:`Command`
            Command to send.

        Returns
        -------
        :class:`Response`
            Parsed response from the adapter.
        """      
        if self.smart_query and self.last_command and command == self.last_command:
            query = ModeAT.REPEAT.build()
        else:
            query = command.build(self.early_return)

        context = Context(command, self.protocol)

        _log.debug(f">>> Send: {str(query)}")

        self.transport.write_bytes(query)
        self.last_command = command

        return self.wait_for_response(context)

    def wait_for_response(self, context: Context) -> Response:
        """
        Wait for a raw response from the transport and parses it using the protocol handler.

        Parameters
        ----------
        context: :class:`Context`
            Context to use for parsing.

        Returns
        -------
        :class:`Response`
            Parsed response or raw fallback response.
        """
        raw = self.transport.read_bytes()

        messages = [
            line
            for line in raw.splitlines()
            if line
        ]

        response_base = ResponseBase(context, raw, messages)

        _log.debug(f"<<< Read:\n{debug_responsebase(response_base)}")

        try:
            return self.protocol_handler.parse_response(response_base)
        except NotImplementedError:
            if self.init_completed:
                _log.warning(f"Unsupported Protocol used: {self.protocol.name}")
            return Response(**vars(response_base))


    def close(self) -> None:
        """
        Closes the transport connection.
        """
        self.transport.close()
        _log.info("Connection closed.")
    

    def __enter__(self):
        """
        Support usage as a context manager.

        Returns
        -------
        :class:`Connection`
            The connection instance itself.
        """
        return self

    def __exit__(self, exc_type: Optional[BaseException], exc_value: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
        """
        Close the connection when exiting the context.
        """
        self.close()


from typing import List

from ..basetypes import BytesRows


class SupportedPIDS():
    def __init__(self, base_pid: int) -> None:
        self.base_pid = base_pid
    
    def __call__(self, parsed_data: BytesRows) -> List[int]:
        concatenated_data = sum(parsed_data, ())

        binary_string = ''.join([f"{int(hex_value, 16):08b}" for hex_value in concatenated_data])

        supported_pids = [
            self.base_pid + i for i, bit in enumerate(binary_string) if bit == "1"
        ]

        return supported_pids



from abc import ABC, abstractmethod


class TransportBase(ABC):
    @abstractmethod
    def connect(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def write_bytes(self, query: bytes) -> None: ...

    @abstractmethod
    def read_bytes(self) -> bytes: ...

    @abstractmethod
    def is_connected(self) -> bool: ...

    def __repr__(self) -> str:
        attrs = ", ".join(f"{k}={v!r}" for k, v in vars(self).items())
        return f"<{self.__class__.__name__}({attrs})>"


from functools import partial

from .group_commands import GroupCommands

from ..command import Command
from ..mode import Mode
from ..parsers.formula import Formula, MultiFormula
from ..parsers.pids import SupportedPIDS


M = Mode.REQUEST
C = partial(Command, M)

F = Formula
MF = MultiFormula
SP = SupportedPIDS

# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_01_-_Show_current_data

class Mode01(GroupCommands):
    """Request Commands - OBD Mode 01 PIDs

    Abbreviations:
        ABS = Anti-lock Braking System
        ALT = Alternative
        AUX = Auxiliary
        DTC = Diagnostic Trouble Code
        EGR = Exhaust Gas Recirculation
        EVAP = Evaporative System # Added
        MAF = Mass Air Flow
        MAX = Maximum
        MIL = Malfunction Indicator Lamp
        OBD = On-Board Diagnostics
        PERC = Percentage
        PID = Parameter ID
        TEMP = Temperature
        VAC = Vacuum
    """

    SUPPORTED_PIDS_A = C(0x00, 0x04, "SUPPORTED_PIDS_A", "PIDs supported [$01 - $20]", None, None, None, SP(0x01))
    STATUS_DTC = C(0x01, 0x04, "STATUS_DTC", "Monitor status since DTCs cleared. (Includes MIL status, DTC count, tests)", None, None, None)
    FREEZE_DTC = C(0x02, 0x02, "FREEZE_DTC", "DTC that caused freeze frame storage.", None, None, None)
    FUEL_STATUS = C(0x03, 0x02, "FUEL_STATUS", "Fuel system status", None, None, None)
    ENGINE_LOAD = C(0x04, 0x01, "ENGINE_LOAD", "Calculated engine load", 0, 100, '%', F("100/255*A"))
    ENGINE_COOLANT_TEMP = C(0x05, 0x01, "ENGINE_COOLANT_TEMP", "Engine coolant temperature", -40, 215, "°C", F("A-40"))
    SHORT_FUEL_TRIM_BANK_1 = C(0x06, 0x01, "SHORT_FUEL_TRIM_BANK_1", "Short term fuel trim (STFT)—Bank 1", -100, 99.2, '%', F("100/128*A-100"))
    LONG_FUEL_TRIM_BANK_1 = C(0x07, 0x01, "LONG_FUEL_TRIM_BANK_1", "Long term fuel trim (LTFT)—Bank 1", -100, 99.2, '%', F("100/128*A-100"))
    SHORT_FUEL_TRIM_BANK_2 = C(0x08, 0x01, "SHORT_FUEL_TRIM_BANK_2", "Short term fuel trim (STFT)—Bank 2", -100, 99.2, '%', F("100/128*A-100"))
    LONG_FUEL_TRIM_BANK_2 = C(0x09, 0x01, "LONG_FUEL_TRIM_BANK_2", "Long term fuel trim (LTFT)—Bank 2", -100, 99.2, '%', F("100/128*A-100"))
    FUEL_PRESSURE = C(0x0A, 0x01, "FUEL_PRESSURE", "Fuel pressure (gauge pressure)", 0, 765, "kPa", F("3*A"))
    INTAKE_PRESSURE = C(0x0B, 0x01, "INTAKE_PRESSURE", "Intake manifold absolute pressure", 0, 255, "kPa", F('A'))
    ENGINE_SPEED = C(0x0C, 0x02, "ENGINE_SPEED", "Engine speed", 0, 16383.75, "rpm", F("(256*A+B)/4"))
    VEHICLE_SPEED = C(0x0D, 0x01, "VEHICLE_SPEED", "Vehicle speed", 0, 255, "km/h", F('A'))
    IGNITION_TIMING_ADVANCE = C(0x0E, 0x01, "IGNITION_TIMING_ADVANCE", "Timing advance", -64, 63.5, "° before TDC", F("A/2-64"))
    INTAKE_AIR_TEMP = C(0x0F, 0x01, "INTAKE_AIR_TEMP", "Intake air temperature", -40, 215, "°C", F("A-40"))
    MAF_RATE = C(0x10, 0x02, "MAF_RATE", "Mass air flow sensor (MAF) air flow rate", 0, 655.35, "g/s", F("(256*A+B)/100")) 
    THROTTLE_POSITION = C(0x11, 0x01, "THROTTLE_POSITION", "Throttle position", 0, 100, '%', F("100/255*A"))
    STATUS_SECONDARY_AIR = C(0x12, 0x01, "STATUS_SECONDARY_AIR", "Commanded secondary air status", None, None, None) 
    OXYGEN_SENSORS_2_BANKS = C(0x13, 0x01, "OXYGEN_SENSORS_2_BANKS", "Oxygen sensors present (in 2 banks)", None, None, None) 
    OXYGEN_SENSOR_1 = C(0x14, 0x02, "OXYGEN_SENSOR_1", "Oxygen Sensor 1 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_2 = C(0x15, 0x02, "OXYGEN_SENSOR_2", "Oxygen Sensor 2 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_3 = C(0x16, 0x02, "OXYGEN_SENSOR_3", "Oxygen Sensor 3 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_4 = C(0x17, 0x02, "OXYGEN_SENSOR_4", "Oxygen Sensor 4 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_5 = C(0x18, 0x02, "OXYGEN_SENSOR_5", "Oxygen Sensor 5 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_6 = C(0x19, 0x02, "OXYGEN_SENSOR_6", "Oxygen Sensor 6 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_7 = C(0x1A, 0x02, "OXYGEN_SENSOR_7", "Oxygen Sensor 7 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OXYGEN_SENSOR_8 = C(0x1B, 0x02, "OXYGEN_SENSOR_8", "Oxygen Sensor 8 A: Voltage B: Short term fuel trim", [0, -100], [1.275, 99.2], ['V', '%'], MF("A/200", "100/128*B-100"))
    OBD_STANDARDS = C(0x1C, 0x01, "OBD_STANDARDS", "OBD standards this vehicle conforms to", 1, 250, None)
    OXYGEN_SENSORS_4_BANKS = C(0x1D, 0x01, "OXYGEN_SENSORS_4_BANKS", "Oxygen sensors present (in 4 banks)", None, None, None) 
    STATUS_AUX_INPUT = C(0x1E, 0x01, "STATUS_AUX_INPUT", "Auxiliary input status (e.g. Power Take Off)", None, None, None)
    ENGINE_RUN_TIME = C(0x1F, 0x02, "ENGINE_RUN_TIME", "Run time since engine start", 0, 65535, 's', F("256*A+B"))

    SUPPORTED_PIDS_B = C(0x20, 0x04, "SUPPORTED_PIDS_B", "PIDs supported [$21 - $40]", None, None, None, SP(0x21))
    MIL_DISTANCE = C(0x21, 0x02, "MIL_DISTANCE", "Distance traveled with MIL on", 0, 65535, "km", F("256*A+B"))
    FUEL_RAIL_PRESSURE_VAC = C(0x22, 0x02, "FUEL_RAIL_PRESSURE_VAC", "Fuel Rail Pressure (relative to manifold vacuum)", 0, 5177.265, "kPa", F("0.079*(256*A+B)"))
    FUEL_RAIL_GAUGE_PRESSURE = C(0x23, 0x02, "FUEL_RAIL_GAUGE_PRESSURE", "Fuel Rail Gauge Pressure (diesel, or gasoline direct injection)", 0, 655350, "kPa", F("10*(256*A+B)"))
    OXYGEN_SENSOR_1_LAMBDA_VOLTAGE = C(0x24, 0x04, "OXYGEN_SENSOR_1_LAMBDA_VOLTAGE", "O2 Sensor 1 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_2_LAMBDA_VOLTAGE = C(0x25, 0x04, "OXYGEN_SENSOR_2_LAMBDA_VOLTAGE", "O2 Sensor 2 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_3_LAMBDA_VOLTAGE = C(0x26, 0x04, "OXYGEN_SENSOR_3_LAMBDA_VOLTAGE", "O2 Sensor 3 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_4_LAMBDA_VOLTAGE = C(0x27, 0x04, "OXYGEN_SENSOR_4_LAMBDA_VOLTAGE", "O2 Sensor 4 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_5_LAMBDA_VOLTAGE = C(0x28, 0x04, "OXYGEN_SENSOR_5_LAMBDA_VOLTAGE", "O2 Sensor 5 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_6_LAMBDA_VOLTAGE = C(0x29, 0x04, "OXYGEN_SENSOR_6_LAMBDA_VOLTAGE", "O2 Sensor 6 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_7_LAMBDA_VOLTAGE = C(0x2A, 0x04, "OXYGEN_SENSOR_7_LAMBDA_VOLTAGE", "O2 Sensor 7 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    OXYGEN_SENSOR_8_LAMBDA_VOLTAGE = C(0x2B, 0x04, "OXYGEN_SENSOR_8_LAMBDA_VOLTAGE", "O2 Sensor 8 Equiv. Ratio (Lambda) & Voltage", [0, 0], [2, 8], ["ratio", 'V'], MF("2/65536*(256*A+B)", "8/65536*(256*C+D)"))
    EGR_PERC = C(0x2C, 0x01, "EGR_PERC", "Percentage of EGR valve opening requested", 0, 100, '%', F("100/255*A"))
    EGR_ERROR = C(0x2D, 0x01, "EGR_ERROR", "EGR Error", -100, 99.2, '%', F("100/128*A-100"))
    COMMANDED_EVAP_PURGE = C(0x2E, 0x01, "COMMANDED_EVAP_PURGE", "Commanded evaporative purge", 0, 100, '%', F("100/255*A")) 
    FUEL_LEVEL = C(0x2F, 0x01, "FUEL_LEVEL", "Fuel Level Input", 0, 100, '%', F("100/255*A")) 
    CLEARED_DTC_WARM_UPS = C(0x30, 0x01, "CLEARED_DTC_WARM_UPS", "Warm-ups since codes cleared", 0, 255, None, F('A'))
    CLEARED_DTC_DISTANCE = C(0x31, 0x02, "CLEARED_DTC_DISTANCE", "Distance traveled since codes cleared", 0, 65535, "km", F("256*A+B"))
    EVAP_PRESSURE = C(0x32, 0x02, "EVAP_PRESSURE", "Evap. System Vapor Pressure", -8192, 8191.75, "Pa")
    BAROMETRIC_PRESSURE = C(0x33, 0x01, "BAROMETRIC_PRESSURE", "Absolute Barometric Pressure", 0, 255, "kPa", F('A'))
    OXYGEN_SENSOR_1_LAMBDA_CURRENT = C(0x34, 0x04, "OXYGEN_SENSOR_1_LAMBDA_CURRENT", "O2 Sensor 1 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_2_LAMBDA_CURRENT = C(0x35, 0x04, "OXYGEN_SENSOR_2_LAMBDA_CURRENT", "O2 Sensor 2 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_3_LAMBDA_CURRENT = C(0x36, 0x04, "OXYGEN_SENSOR_3_LAMBDA_CURRENT", "O2 Sensor 3 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_4_LAMBDA_CURRENT = C(0x37, 0x04, "OXYGEN_SENSOR_4_LAMBDA_CURRENT", "O2 Sensor 4 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_5_LAMBDA_CURRENT = C(0x38, 0x04, "OXYGEN_SENSOR_5_LAMBDA_CURRENT", "O2 Sensor 5 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_6_LAMBDA_CURRENT = C(0x39, 0x04, "OXYGEN_SENSOR_6_LAMBDA_CURRENT", "O2 Sensor 6 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_7_LAMBDA_CURRENT = C(0x3A, 0x04, "OXYGEN_SENSOR_7_LAMBDA_CURRENT", "O2 Sensor 7 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    OXYGEN_SENSOR_8_LAMBDA_CURRENT = C(0x3B, 0x04, "OXYGEN_SENSOR_8_LAMBDA_CURRENT", "O2 Sensor 8 Equiv. Ratio (Lambda) & Current", [0, -128], [2, 128], ["ratio", "mA"], MF("2/65536*(256*A+B)", "(256*C+D)/256-128"))
    CATALYST_TEMP_BANK_1_SENSOR_1 = C(0x3C, 0x02, "CATALYST_TEMP_BANK_1_SENSOR_1", "Catalyst Temperature: Bank 1, Sensor 1", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_BANK_2_SENSOR_1 = C(0x3D, 0x02, "CATALYST_TEMP_BANK_2_SENSOR_1", "Catalyst Temperature: Bank 2, Sensor 1", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_BANK_1_SENSOR_2 = C(0x3E, 0x02, "CATALYST_TEMP_BANK_1_SENSOR_2", "Catalyst Temperature: Bank 1, Sensor 2", -40, 6513.5, "°C", F("(256*A+B)/10-40"))
    CATALYST_TEMP_BANK_2_SENSOR_2 = C(0x3F, 0x02, "CATALYST_TEMP_BANK_2_SENSOR_2", "Catalyst Temperature: Bank 2, Sensor 2", -40, 6513.5, "°C", F("(256*A+B)/10-40"))

    SUPPORTED_PIDS_C = C(0x40, 0x04, "SUPPORTED_PIDS_C", "PIDs supported [$41 - $60]", None, None, None, SP(0x41))
    STATUS_DRIVE_CYCLE = C(0x41, 0x04, "STATUS_DRIVE_CYCLE", "Monitor status this drive cycle", None, None, None)
    VEHICLE_VOLTAGE = C(0x42, 0x02, "VEHICLE_VOLTAGE", "Control module voltage", 0, 65.535, 'V', F("(256*A+B)/1000"))
    ENGINE_LOAD_ABSOLUTE = C(0x43, 0x02, "ENGINE_LOAD_ABSOLUTE", "Absolute percentage calculated from air mass intake", 0, 25700, '%', F("100/255*(256*A+B)"))
    COMMANDED_AIR_FUEL_RATIO = C(0x44, 0x02, "COMMANDED_AIR_FUEL_RATIO", "Commanded Air-Fuel Equivalence Ratio (lambda,λ)", 0, 2, "ratio", F("2/65536*(256*A+B)")) 
    THROTTLE_POSITION_RELATIVE = C(0x45, 0x01, "THROTTLE_POSITION_RELATIVE", "Relative throttle position", 0, 100, '%', F("100/255*A"))
    AMBIENT_AIR_TEMP = C(0x46, 0x01, "AMBIENT_AIR_TEMP", "Ambient air temperature", -40, 215, "°C", F("A-40"))
    THROTTLE_POSITION_B = C(0x47, 0x01, "THROTTLE_POSITION_B", "Absolute throttle position B", 0, 100, '%', F("100/255*A"))
    THROTTLE_POSITION_C = C(0x48, 0x01, "THROTTLE_POSITION_C", "Absolute throttle position C", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_D = C(0x49, 0x01, "ACCELERATOR_POSITION_D", "Accelerator pedal position D", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_E = C(0x4A, 0x01, "ACCELERATOR_POSITION_E", "Accelerator pedal position E", 0, 100, '%', F("100/255*A"))
    ACCELERATOR_POSITION_F = C(0x4B, 0x01, "ACCELERATOR_POSITION_F", "Accelerator pedal position F", 0, 100, '%', F("100/255*A"))
    THROTTLE_ACTUATOR = C(0x4C, 0x01, "THROTTLE_ACTUATOR", "Commanded throttle actuator", 0, 100, '%', F("100/255*A"))
    MIL_RUN_TIME = C(0x4D, 0x02, "MIL_RUN_TIME", "Time run with MIL on", 0, 65535, "min", F("256*A+B"))
    CLEARED_DTC_SINCE = C(0x4E, 0x02, "CLEARED_DTC_SINCE", "Time since trouble codes cleared", 0, 65535, "min", F("256*A+B"))
    MAX_FUEL_AIR_RATIO_O2_VOLT_CURR_PRESSURE = C(0x4F, 0x04, "MAX_FUEL_AIR_RATIO_O2_VOLT_CURR_PRESSURE", "Maximum value for Equiv Ratio, O2 Sensor V, O2 Sensor I, Intake Pressure", [0, 0, 0, 0], [255, 255, 255, 2550], ["ratio", 'V', "mA", "kPa"], MF('A', 'B', 'C', 'D*10'))
    MAF_MAX = C(0x50, 0x04, "MAF_MAX", "Maximum value for MAF rate", 0, 2550, "g/s", F("A*10"))
    FUEL_TYPE = C(0x51, 0x01, "FUEL_TYPE", "Fuel Type", None, None, None)
    ETHANOL_PERC = C(0x52, 0x01, "ETHANOL_PERC", "Ethanol fuel %", 0, 100, '%', F("100/255*A"))
    EVAP_PRESSURE_ABSOLUTE = C(0x53, 0x02, "EVAP_PRESSURE_ABSOLUTE", "Absolute Evap system Vapor Pressure", 0, 327.675, "kPa", F("(256*A+B)/200")) 
    EVAP_PRESSURE_ALT = C(0x54, 0x02, "EVAP_PRESSURE_ALT", "Evap system vapor pressure (alternate encoding)", -32768, 32767, "Pa")
    SHORT_OXYGEN_TRIM_BANK_1 = C(0x55, 0x02, "SHORT_OXYGEN_TRIM_BANK_1", "Short term secondary O2 sensor trim, A: bank 1, B: bank 3", -100, 99.2, '%', MF("100/128*A-100", "100/128*B-100"))
    LONG_OXYGEN_TRIM_BANK_1 = C(0x56, 0x02, "LONG_OXYGEN_TRIM_BANK_1", "Long term secondary O2 sensor trim, A: bank 1, B: bank 3", -100, 99.2, '%', MF("100/128*A-100", "100/128*B-100"))
    SHORT_OXYGEN_TRIM_BANK_2 = C(0x57, 0x02, "SHORT_OXYGEN_TRIM_BANK_2", "Short term secondary O2 sensor trim, A: bank 2, B: bank 4", -100, 99.2, '%', MF("100/128*A-100", "100/128*B-100"))
    LONG_OXYGEN_TRIM_BANK_2 = C(0x58, 0x02, "LONG_OXYGEN_TRIM_BANK_2", "Long term secondary O2 sensor trim, A: bank 2, B: bank 4", -100, 99.2, '%', MF("100/128*A-100", "100/128*B-100"))
    FUEL_RAIL_PRESSURE = C(0x59, 0x02, "FUEL_RAIL_PRESSURE", "Fuel rail absolute pressure", 0, 655350, "kPa", F("10*(256*A+B)"))
    ACCELERATOR_POSITION_RELATIVE = C(0x5A, 0x01, "ACCELERATOR_POSITION_RELATIVE", "Relative accelerator pedal position", 0, 100, '%', F("100/255*A"))
    HYBRID_BATTERY_REMAINING = C(0x5B, 0x01, "HYBRID_BATTERY_REMAINING", "Hybrid battery pack remaining life", 0, 100, '%', F("100/255*A"))
    ENGINE_OIL_TEMP = C(0x5C, 0x01, "ENGINE_OIL_TEMP", "Engine oil temperature", -40, 210, "°C", F("A-40"))
    FUEL_INJECTION_TIMING = C(0x5D, 0x02, "FUEL_INJECTION_TIMING", "Fuel injection timing", -210.00, 301.992, "°", F("(256*A+B)/128-210"))
    ENGINE_FUEL_RATE = C(0x5E, 0x02, "ENGINE_FUEL_RATE", "Engine fuel rate", 0, 3212.75, "L/h", F("(256*A+B)/20"))
    VEHICLE_EMISSION_STANDARDS = C(0x5F, 0x01, "VEHICLE_EMISSION_STANDARDS", "Emission requirements to which vehicle is designed", None, None, None) 

    SUPPORTED_PIDS_D = C(0x60, 0x04, "SUPPORTED_PIDS_D", "PIDs supported [$61 - $80]", None, None, None, SP(0x61))
    # ENGINE_TORQUE_DEMAND = C(0x61, 0x01, "ENGINE_TORQUE_DEMAND", "Driver's demand engine percent torque", -125, 130, '%', F("A-125"))
    # ENGINE_TORQUE = C(0x62, 0x01, "ENGINE_TORQUE", "Actual engine percent torque", -125, 130, '%', F("A-125"))
    # ENGINE_TORQUE_REF = C(0x63, 0x02, "ENGINE_TORQUE_REF", "Engine reference torque", 0, 65535, "N⋅m", F("256*A+B"))
    # ENGINE_TORQUE_DATA = C(0x64, 0x05, "ENGINE_TORQUE_DATA", "Engine percent torque data", -125, 130, '%', MF("A-125", "B-125", "C-125", "D-125", "E-125"))
    # AUXILIARY_INPUT_OUTPUT_SUPPORTED = C(0x65, 0x02, "AUXILIARY_INPUT_OUTPUT_SUPPORTED", "Auxiliary input / output supported", None, None, None)
    # MAF_SENSOR = C(0x66, 0x05, "MAF_SENSOR", "Mass air flow sensor", 0, 2047.96875, "g/s")
    # ENGINE_COOLANT_TEMP = C(0x67, 0x03, "ENGINE_COOLANT_TEMP", "Engine coolant temperature", -40, 215, "°C")
    # INTAKE_AIR_TEMP_SENSOR = C(0x68, 0x03, "INTAKE_AIR_TEMP_SENSOR", "Intake air temperature sensor", -40, 215, "°C")
    # EGR_ACTUAL = C(0x69, 0x07, "ACTUAL_EGR", "Actual EGR", None, None, None)
# COMMANDED_DIESEL_INTAKE_AIR_FLOW_CONTROL_AND_RELATIVE_INTAKE_AIR_FLOW_POSITION = C(0x6A, 0x05, "Commanded Diesel intake air flow control and relative intake air flow position", "Commanded Diesel intake air flow control and relative intake air flow position", None, None, None)
# EXHAUST_GAS_RECIRCULATION_TEMPERATURE = C(0x6B, 0x05, "Exhaust gas recirculation temperature", "Exhaust gas recirculation temperature", None, None, None)
# COMMANDED_THROTTLE_ACTUATOR_CONTROL_AND_RELATIVE_THROTTLE_POSITION = C(0x6C, 0x05, "Commanded throttle actuator control and relative throttle position", "Commanded throttle actuator control and relative throttle position", None, None, None)
# FUEL_PRESSURE_CONTROL_SYSTEM = C(0x6D, 0x0B, "Fuel pressure control system", "Fuel pressure control system", None, None, None)
# INJECTION_PRESSURE_CONTROL_SYSTEM = C(0x6E, 0x09, "Injection pressure control system", "Injection pressure control system", None, None, None)
# TURBOCHARGER_COMPRESSOR_INLET_PRESSURE = C(0x6F, 0x03, "Turbocharger compressor inlet pressure", "Turbocharger compressor inlet pressure", None, None, None)
# BOOST_PRESSURE_CONTROL = C(0x70, 0x0A, "Boost pressure control", "Boost pressure control", None, None, None)
# VARIABLE_GEOMETRY_TURBO_(VGT)_CONTROL = C(0x71, 0x06, "Variable Geometry turbo (VGT) control", "Variable Geometry turbo (VGT) control", None, None, None)
# WASTEGATE_CONTROL = C(0x72, 0x05, "Wastegate control", "Wastegate control", None, None, None)
# EXHAUST_PRESSURE = C(0x73, 0x05, "Exhaust pressure", "Exhaust pressure", None, None, None)
# TURBOCHARGER_RPM = C(0x74, 0x05, "Turbocharger RPM", "Turbocharger RPM", None, None, None)
# TURBOCHARGER_TEMPERATURE = C(0x75, 0x07, "Turbocharger temperature", "Turbocharger temperature", None, None, None)
# TURBOCHARGER_TEMPERATURE = C(0x76, 0x07, "Turbocharger temperature", "Turbocharger temperature", None, None, None)
# CHARGE_AIR_COOLER_TEMPERATURE_(CACT) = C(0x77, 0x05, "Charge air cooler temperature (CACT)", "Charge air cooler temperature (CACT)", None, None, None)
# EXHAUST_GAS_TEMPERATURE_(EGT)_BANK_1 = C(0x78, 0x09, "Exhaust Gas temperature (EGT) Bank 1", "Exhaust Gas temperature (EGT) Bank 1", None, None, None)
# EXHAUST_GAS_TEMPERATURE_(EGT)_BANK_2 = C(0x79, 0x09, "Exhaust Gas temperature (EGT) Bank 2", "Exhaust Gas temperature (EGT) Bank 2", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF)_DIFFERENTIAL_PRESSURE = C(0x7A, 0x07, "Diesel particulate filter (DPF) differential pressure", "Diesel particulate filter (DPF) differential pressure", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF) = C(0x7B, 0x07, "Diesel particulate filter (DPF)", "Diesel particulate filter (DPF)", None, None, None)
# DIESEL_PARTICULATE_FILTER_(DPF)_TEMPERATURE = C(0x7C, 0x09, "Diesel Particulate filter (DPF) temperature", "Diesel Particulate filter (DPF) temperature", None, None, "°C")
# NOX_NTE_(NOT-TO-EXCEED)_CONTROL_AREA_STATUS = C(0x7D, 0x01, "NOx NTE (Not-To-Exceed) control area status", "NOx NTE (Not-To-Exceed) control area status", None, None, None)
# PM_NTE_(NOT-TO-EXCEED)_CONTROL_AREA_STATUS = C(0x7E, 0x01, "PM NTE (Not-To-Exceed) control area status", "PM NTE (Not-To-Exceed) control area status", None, None, None)
# ENGINE_RUN_TIME_[B] = C(0x7F, 0x0D, "Engine run time [b]", "Engine run time [b]", None, None, "s")

    SUPPORTED_PIDS_E = C(0x80, 0x04, "SUPPORTED_PIDS_E", "PIDs supported [$81 - $A0]", None, None, None, SP(0x81))
# ENGINE_RUN_TIME_FOR_AUXILIARY_EMISSIONS_CONTROL_DEVICE(AECD) = C(0x81, 0x29, "Engine run time for Auxiliary Emissions Control Device(AECD)", "Engine run time for Auxiliary Emissions Control Device(AECD)", None, None, None)
# ENGINE_RUN_TIME_FOR_AUXILIARY_EMISSIONS_CONTROL_DEVICE(AECD) = C(0x82, 0x29, "Engine run time for Auxiliary Emissions Control Device(AECD)", "Engine run time for Auxiliary Emissions Control Device(AECD)", None, None, None)
# NOX_SENSOR = C(0x83, 0x09, "NOx sensor", "NOx sensor", None, None, None)
# MANIFOLD_SURFACE_TEMPERATURE = C(0x84, 0x01, "Manifold surface temperature", "Manifold surface temperature", None, None, None)
# NOX_REAGENT_SYSTEM = C(0x85, 0x0A, "NOx reagent system", "NOx reagent system", None, None, '%')
# PARTICULATE_MATTER_(PM)_SENSOR = C(0x86, 0x05, "Particulate matter (PM) sensor", "Particulate matter (PM) sensor", None, None, None)
# INTAKE_MANIFOLD_ABSOLUTE_PRESSURE = C(0x87, 0x05, "Intake manifold absolute pressure", "Intake manifold absolute pressure", None, None, None)
# SCR_INDUCE_SYSTEM = C(0x88, 0x0D, "SCR Induce System", "SCR Induce System", None, None, None)
# RUN_TIME_FOR_AECD_#11-#15 = C(0x89, 0x29, "Run Time for AECD #11-#15", "Run Time for AECD #11-#15", None, None, None)
# RUN_TIME_FOR_AECD_#16-#20 = C(0x8A, 0x29, "Run Time for AECD #16-#20", "Run Time for AECD #16-#20", None, None, None)
# DIESEL_AFTERTREATMENT = C(0x8B, 0x07, "Diesel Aftertreatment", "Diesel Aftertreatment", None, None, None)
# O2_SENSOR_(WIDE_RANGE) = C(0x8C, 0x11, "O2 Sensor (Wide Range)", "O2 Sensor (Wide Range)", None, None, None)
# THROTTLE_POSITION_G = C(0x8D, 0x01, "Throttle Position G", "Throttle Position G", 0, 100, '%')
# ENGINE_FRICTION_-_PERCENT_TORQUE = C(0x8E, 0x01, "Engine Friction - Percent Torque", "Engine Friction - Percent Torque", -125, 130, '%')
# PM_SENSOR_BANK_1_&_2 = C(0x8F, 0x07, "PM Sensor Bank 1 & 2", "PM Sensor Bank 1 & 2", None, None, None)
# WWH-OBD_VEHICLE_OBD_SYSTEM_INFORMATION = C(0x90, 0x03, "WWH-OBD Vehicle OBD System Information", "WWH-OBD Vehicle OBD System Information", None, None, "h")
# WWH-OBD_VEHICLE_OBD_SYSTEM_INFORMATION = C(0x91, 0x05, "WWH-OBD Vehicle OBD System Information", "WWH-OBD Vehicle OBD System Information", None, None, "h")
# FUEL_SYSTEM_CONTROL = C(0x92, 0x02, "Fuel System Control", "Fuel System Control", None, None, None)
# WWH-OBD_VEHICLE_OBD_COUNTERS_SUPPORT = C(0x93, 0x03, "WWH-OBD Vehicle OBD Counters support", "WWH-OBD Vehicle OBD Counters support", None, None, "h")
# NOX_WARNING_AND_INDUCEMENT_SYSTEM = C(0x94, 0x0C, "NOx Warning And Inducement System", "NOx Warning And Inducement System", None, None, None)
# EXHAUST_GAS_TEMPERATURE_SENSOR = C(0x98, 0x09, "Exhaust Gas Temperature Sensor", "Exhaust Gas Temperature Sensor", None, None, None)
# EXHAUST_GAS_TEMPERATURE_SENSOR = C(0x99, 0x09, "Exhaust Gas Temperature Sensor", "Exhaust Gas Temperature Sensor", None, None, None)
# HYBRID/EV_VEHICLE_SYSTEM_DATA,_BATTERY,_VOLTAGE = C(0x9A, 0x06, "Hybrid/EV Vehicle System Data, Battery, Voltage", "Hybrid/EV Vehicle System Data, Battery, Voltage", None, None, None)
# DIESEL_EXHAUST_FLUID_SENSOR_DATA = C(0x9B, 0x04, "Diesel Exhaust Fluid Sensor Data", "Diesel Exhaust Fluid Sensor Data", None, None, '%')
# O2_SENSOR_DATA = C(0x9C, 0x11, "O2 Sensor Data", "O2 Sensor Data", None, None, None)
# ENGINE_FUEL_RATE = C(0x9D, 0x04, "Engine Fuel Rate", "Engine Fuel Rate", None, None, "g/s")
# ENGINE_EXHAUST_FLOW_RATE = C(0x9E, 0x02, "Engine Exhaust Flow Rate", "Engine Exhaust Flow Rate", None, None, "kg/h")
# FUEL_SYSTEM_PERCENTAGE_USE = C(0x9F, 0x09, "Fuel System Percentage Use", "Fuel System Percentage Use", None, None, None)

    SUPPORTED_PIDS_F = C(0xA0, 0x04, "SUPPORTED_PIDS_F", "PIDs supported [$A1 - $C0]", None, None, None, SP(0xA1))
# NOX_SENSOR_CORRECTED_DATA = C(0xA1, 0x09, "NOx Sensor Corrected Data", "NOx Sensor Corrected Data", None, None, "ppm")
# CYLINDER_FUEL_RATE = C(0xA2, 0x02, "Cylinder Fuel Rate", "Cylinder Fuel Rate", 0, 2047.96875, "mg/stroke")
# EVAP_SYSTEM_VAPOR_PRESSURE = C(0xA3, 0x09, "Evap System Vapor Pressure", "Evap System Vapor Pressure", None, None, "Pa")
# TRANSMISSION_ACTUAL_GEAR = C(0xA4, 0x04, "Transmission Actual Gear", "Transmission Actual Gear", 0, 65.535, "ratio")
# COMMANDED_DIESEL_EXHAUST_FLUID_DOSING = C(0xA5, 0x04, "Commanded Diesel Exhaust Fluid Dosing", "Commanded Diesel Exhaust Fluid Dosing", 0, 127.5, '%')
# ODOMETER_[C] = C(0xA6, 0x04, "Odometer [c]", "Odometer [c]", 0, 429496729.5, "km")
# NOX_SENSOR_CONCENTRATION_SENSORS_3_AND_4 = C(0xA7, 0x04, "NOx Sensor Concentration Sensors 3 and 4", "NOx Sensor Concentration Sensors 3 and 4", None, None, None)
# NOX_SENSOR_CORRECTED_CONCENTRATION_SENSORS_3_AND_4 = C(0xA8, 0x04, "NOx Sensor Corrected Concentration Sensors 3 and 4", "NOx Sensor Corrected Concentration Sensors 3 and 4", None, None, None)
# ABS_DISABLE_SWITCH_STATE = C(0xA9, 0x04, "ABS Disable Switch State", "ABS Disable Switch State", None, None, None)

    SUPPORTED_PIDS_G = C(0xC0, 0x04, "SUPPORTED_PIDS_G", "PIDs supported [$C1 - $E0]", None, None, None, SP(0xC1))
# FUEL_LEVEL_INPUT_A/B = C(0xC3, 0x02, "Fuel Level Input A/B", "Fuel Level Input A/B", 0, 25700, '%')
# EXHAUST_PARTICULATE_CONTROL_SYSTEM_DIAGNOSTIC_TIME/COUNT = C(0xC4, 0x08, "Exhaust Particulate Control System Diagnostic Time/Count", "Exhaust Particulate Control System Diagnostic Time/Count", 0, 4294967295, "seconds / Count")
# FUEL_PRESSURE_A_AND_B = C(0xC5, 0x04, "Fuel Pressure A and B", "Fuel Pressure A and B", 0, 5177, "kPa")
# BYTE_1_-_PARTICULATE_CONTROL_-_DRIVER_INDUCEMENT_SYSTEM_STATUS_BYTE_2,3_-_REMOVAL_OR_BLOCK_OF_THE_PARTICULATE_AFTERTREATMENT_SYSTEM_COUNTER_BYTE_4,5_-_LIQUID_REGENT_INJECTION_SYSTEM_(E.G._FUEL-BORNE_CATALYST)_FAILURE_COUNTER_BYTE_6,7_-_MALFUNCTION_OF_PARTICULATE_CONTROL_MONITORING_SYSTEM_COUNTER = C(0xC6, 0x07, "Byte 1 - Particulate control - driver inducement system status Byte 2,3 - Removal or block of the particulate aftertreatment system counter Byte 4,5 - Liquid regent injection system (e.g. fuel-borne catalyst) failure counter Byte 6,7 - Malfunction of Particulate control monitoring system counter", "Byte 1 - Particulate control - driver inducement system status Byte 2,3 - Removal or block of the particulate aftertreatment system counter Byte 4,5 - Liquid regent injection system (e.g. fuel-borne catalyst) failure counter Byte 6,7 - Malfunction of Particulate control monitoring system counter", 0, 65535, "h")
# DISTANCE_SINCE_REFLASH_OR_MODULE_REPLACEMENT = C(0xC7, 0x02, "Distance Since Reflash or Module Replacement", "Distance Since Reflash or Module Replacement", 0, 65535, "km")
# NOX_CONTROL_DIAGNOSTIC_(NCD)_AND_PARTICULATE_CONTROL_DIAGNOSTIC_(PCD)_WARNING_LAMP_STATUS = C(0xC8, 0x01, "NOx Control Diagnostic (NCD) and Particulate Control Diagnostic (PCD) Warning Lamp status", "NOx Control Diagnostic (NCD) and Particulate Control Diagnostic (PCD) Warning Lamp status", -, -, "Bit")


from functools import partial

from .group_commands import GroupCommands

from ..command import Command
from ..mode import Mode


M = Mode.STATUS_DTC
C = partial(Command, M)

# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_03_-_Show_stored_Diagnostic_Trouble_Codes_(DTCs)

class Mode03(GroupCommands):
    """Get Diagnostic Trouble Codes Command"""
    GET_DTC = C('', 0x00, "GET_DTC", "Request trouble codes", None, None, None)


import pytest

from typing import Any, Dict

from obdii.command import Command
from obdii.mode import Mode


@pytest.fixture
def arg_command_factory():
    """Fixture that returns a function to create Command instances."""
    def _create_command(pid: str, cmd_args: Dict[str, Any]) -> Command:
        return Command(
            mode=Mode.AT,
            pid=pid,
            n_bytes=0x02,
            name="Test Command",
            description="A command for testing",
            min_values=0,
            max_values=255,
            units='V',
            formula=None,
            command_args=cmd_args,
        )
    return _create_command


@pytest.mark.parametrize(
    "pid, command_args, arguments, expected_pid",
    [
        # Basic formatting tests
        ("TEST {h}", {'h': int}, [0], "TEST 0"),
        ("TEST {x} TEST {y}", {'x': int, 'y': int}, [0, 1], "TEST 0 TEST 1"),
        ("TEST {h}", {'h': int}, [10], "TEST A"),
        ("TEST {x} TEST {y}", {'x': int, 'y': int}, [10, 11], "TEST A TEST B"),
        
        # Hexadecimal formatting tests
        ("TEST {hh}", {"hh": int}, [0], "TEST 00"),
        ("TEST {xx} TEST {yy}", {"xx": int, "yy": int}, [0, 1], "TEST 00 TEST 01"),
        ("TEST {hh}", {"hh": int}, [10], "TEST 0A"),
        ("TEST {xx} TEST {yy}", {"xx": int, "yy": int}, [10, 11], "TEST 0A TEST 0B"),
        ("TEST {hh}", {"hh": int}, [17], "TEST 11"),
        ("TEST {xx} TEST {yy}", {"xx": int, "yy": int}, [17, 18], "TEST 11 TEST 12"),
        
        # Edge cases
        ("TEST {hh}", {"hh": int}, [255], "TEST FF"),
    ]
)
def test_command_correct_arguments(arg_command_factory, pid, command_args, arguments, expected_pid):
    command = arg_command_factory(pid, command_args)

    called_command = command(*arguments)

    assert called_command.is_formatted == True
    assert called_command.pid == expected_pid
    assert called_command.pid != command.pid


@pytest.mark.parametrize(
    "pid, command_args, arguments, expected_exception",
    [
        # Missing arguments
        ("TEST {h}", {'h': int}, [], ValueError),
        ("TEST {x} TEST {y}", {'x': int, 'y': int}, [1], ValueError),

        # Too many arguments
        ("TEST {x}", {'x': int}, [1, 2], ValueError),

        # Incorrect types
        ("TEST {h}", {'h': int}, ["string"], TypeError),
        ("TEST {xx}", {"xx": str}, [1], TypeError),
        ("TEST {xx}", {"xx": str}, [1.5], TypeError),

        # Argument formatting issues
        ("TEST {hh}", {"hh": int}, [256], ValueError),
        ("TEST {xx}", {"xx": str}, ["string"], ValueError),
        ("TEST {xx}", {"xx": str}, ["s"], ValueError),

        # Missing placeholders
        ("TEST", {'h': int}, [1], ValueError),
        ("TEST", {"xx": str, "yy": str}, ["ii", "jj"], ValueError),

        # Edge Cases
        ("TEST", {}, [], ValueError),
        ("TEST {h}", {}, [], ValueError),
        ("TEST {h}", {}, [1], ValueError),
        ("TEST {h}", {'h': int}, [None], TypeError),
        ("TEST {xx}", {"xx": str}, [None], TypeError),
        ("TEST {xx}", {"xx": str}, [''], ValueError),
        ("TEST {hh}", {"hh": int}, [-1], ValueError),
    ]
)
def test_command_invalid_arguments(arg_command_factory, pid, command_args, arguments, expected_exception):
    command = arg_command_factory(pid, command_args)

    with pytest.raises(expected_exception):
        command(*arguments)


@pytest.mark.parametrize(
    "pid, command_args, expected_query",
    [
        # 1. Basic Command
        (0x01, None, b"AT 01\r"),
        (0x01, {}, b"AT 01\r"),
        (0x0A, None, b"AT 0A\r"),
        (0xFF, {}, b"AT FF\r"),
        ("01", None, b"AT 01\r"),
        ("01", {}, b"AT 01\r"),
        ("TEST", None, b"AT TEST\r"),
        ("TEST", {}, b"AT TEST\r"),

        # 2. Missing Arguments (and Expected Errors)
        (0x01, {'h': int}, ValueError),
        ("01", {'h': int}, ValueError),
    ]
)
def test_build_function(arg_command_factory, pid, command_args, expected_query):
    command = arg_command_factory(pid, command_args)

    if isinstance(expected_query, bytes):
        query = command.build()
        assert query == expected_query
    else:
        with pytest.raises(expected_query):
            command.build()


from os import environ
from setuptools import setup

ci_environment = environ.get("CI", "false").lower() == "true"

local_scheme = "no-local-version" if ci_environment else "node-and-date"

setup(
    use_scm_version={
        "local_scheme": local_scheme,
    }
)


from __future__ import annotations

from re import compile, Pattern
from typing import Set, Type, Optional


class ResponseBaseError(Exception):
    _registry: Set[Type[ResponseBaseError]] = set()

    pattern: Optional[bytes] = None
    regex_pattern: Optional[Pattern] = None

    def __init__(self, response: bytes) -> None:
        self.response = response
        self.message = f"{self.__class__.__name__}: {self.__class__.__doc__} - Got: {str(self.response)}"

        super().__init__(self.message)

    def __init_subclass__(cls) -> None:
        if not cls.pattern and not cls.regex_pattern:
            raise TypeError(f"{cls.__name__} must define either 'pattern' or 'regex_pattern'")
        ResponseBaseError._registry.add(cls)

    @classmethod
    def detect(cls, response: bytes) -> Optional[ResponseBaseError]:
        """Detect an error in a response and return the corresponding error instance."""
        for err in cls._registry:
            if err.pattern and err.pattern in response:
                return err(response)

            if err.regex_pattern and err.regex_pattern.search(response):
                return err(response)

        return None


# Errors

class InvalidCommandError(ResponseBaseError):
    """The command received on the RS232 input was not recognized or misunderstood."""
    pattern = b'?'

class BufferFullError(ResponseBaseError):
    """The ELM327's 256-byte buffer is full."""
    pattern = b"BUFFER FULL"

class BusBusyError(ResponseBaseError):
    """The ELM327 detected excessive bus activity and was unable to respond."""
    pattern = b"BUS BUSY"

class BusError(ResponseBaseError):
    """A generic error has occured on the bus."""
    pattern = b"BUS ERROR"

class CanError(ResponseBaseError):
    """The CAN system had difficulty initializing, sending, or receiving."""
    pattern = b"CAN ERROR"

class InvalidDataError(ResponseBaseError):
    """There was a response from the vehicle, but the information was incorrect or could not be recovered."""
    regex_pattern = compile(rb"(?<!<)DATA ERROR")

class InvalidLineError(ResponseBaseError):
    """There was an error in the line that this points to."""
    pattern = b"<DATA ERROR"

class DeviceInternalError(ResponseBaseError):
    """Internal errors reported as ERR with a two digit code following."""
    regex_pattern = compile(rb"ERR\d{2}")

class SignalFeedbackError(ResponseBaseError):
    """Output energized, but input signal not detected. Possible wiring issue. Verify connections."""
    pattern = b"FB ERROR"

class MissingDataError(ResponseBaseError):
    """Vehicle did not respond within timeout."""
    pattern = b"NO DATA"

class CanDataError(ResponseBaseError):
    """Received CAN data contains errors. Verify protocol and baud rate settings."""
    pattern = b"<RX ERROR"

class StoppedError(ResponseBaseError):
    """Operation interrupted by RS232 character or low RTS signal."""
    pattern = b"STOPPED"

class ProtocolConnectionError(ResponseBaseError):
    """No supported protocol detected. Verify vehicle ignition status, compatibility, and connections."""
    pattern = b"UNABLE TO CONNECT"


# Warnings

class InactivityWarning(ResponseBaseError):
    """No RS232 activity has been detected for a specified duration (4 or 19 minutes)."""
    pattern = b"ACT ALERT"

class LowPowerWarning(ResponseBaseError):
    """The ELM327 is entering Low Power mode in 2 seconds. No interruption possible."""
    pattern = b"LP ALERT"

class LowVoltageResetWarning(ResponseBaseError):
    """Indicates low 5V supply voltage, triggering a reset."""
    pattern = b"LV RESET"


from importlib.metadata import version, PackageNotFoundError
from logging import NullHandler, getLogger
from pkgutil import extend_path

from .command import Command
from .connection import Connection
from .mode import Mode
from .modes import at_commands, commands
from .protocol import Protocol
# note: Need to import the protocols to register them
from .protocols import ProtocolCAN, ProtocolJ1850, ProtocolKWP2000
from .response import ResponseBase, Context, Response


__title__ = "obdii"
__author__ = "PaulMarisOUMary"
__license__ = "MIT"
__copyright__ = "Copyright 2025-present PaulMarisOUMary"

try:
    __version__ = version("py-obdii")
except PackageNotFoundError:
    __version__ = "0.0.0"
__path__ = extend_path(__path__, __name__)


__all__ = [
    "at_commands",
    "commands",
    "Command",
    "Connection",
    "Context",
    "Mode",
    "Protocol",
    "Response",
    "ResponseBase",
]

getLogger(__name__).addHandler(NullHandler())


import pytest

from obdii.basetypes import _MissingSentinel, MISSING


missing_test_cases = [
    (_MissingSentinel(), True),
    (object(), False),
    (42, False),
]


@pytest.mark.parametrize("obj, expected", missing_test_cases)
def test_missing_eq_operator(obj, expected):
    assert (MISSING == obj) is expected, f"Expected MISSING == {obj} to be {expected}"


@pytest.mark.parametrize("obj, expected", missing_test_cases)
def test_missing_is_operator(obj, expected):
    assert (MISSING is obj) is expected, f"Expected MISSING is {obj} to be {expected}"


def test_missing_singleton_behavior():
    assert MISSING is _MissingSentinel(), "MISSING is not the singleton instance"

    missing1 = _MissingSentinel()
    missing2 = _MissingSentinel()

    assert missing1 is missing2, "MISSING instances are not the same"


@pytest.mark.parametrize("obj, expected_repr", [
    (MISSING, "..."),
    (_MissingSentinel(), "..."),
])
def test_missing_repr(obj, expected_repr):
    assert repr(obj) == expected_repr, f"Expected repr({obj}) to be '{expected_repr}'"


@pytest.mark.parametrize("obj, expected_bool", [
    (MISSING, False),
    (_MissingSentinel(), False),
])
def test_missing_bool(obj, expected_bool):
    assert bool(obj) is expected_bool, f"Expected bool({obj}) to be {expected_bool}"


@pytest.mark.parametrize("obj, expected_hash", [
    (MISSING, 0),
    (_MissingSentinel(), 0),
])
def test_missing_hash(obj, expected_hash):
    assert hash(obj) == expected_hash, f"Expected hash({obj}) to be {expected_hash}"


from functools import partial

from .group_commands import GroupCommands

from ..command import Command
from ..mode import Mode
from ..parsers.pids import SupportedPIDS


M = Mode.VEHICLE_INFO
C = partial(Command, M)
SP = SupportedPIDS

# https://en.wikipedia.org/wiki/OBD-II_PIDs#Service_09_-_Request_vehicle_information

class Mode09(GroupCommands):
    """Request Vehicle Information"""

    SUPPORTED_PIDS_9 = C(0x00, 0x04, "SUPPORTED_PIDS_9", "Service 9 supported PIDs [$01 to $20]", None, None, None, SP(0x01))
    VIN_MESSAGE_COUNT = C(0x01, 0x01, "VIN_MESSAGE_COUNT", "VIN Message Count in PID 02. Only for ISO 9141-2, ISO 14230-4 and SAE J1850.", None, None, None)
    VIN = C(0x02, 0x11, "VIN", "Vehicle Identification Number (VIN)", None, None, None)
    CALIBRATION_ID_MESSAGE_COUNT = C(0x03, 0x01, "CALIBRATION_ID_MESSAGE_COUNT", "Calibration ID message count for PID 04. Only for ISO 9141-2, ISO 14230-4 and SAE J1850.", None, None, None)
    CALIBRATION_ID = C(0x04, 0x00, "CALIBRATION_ID", "Calibration ID", None, None, None)
    CVN_MESSAGE_COUNT = C(0x05, 1, "CVN_MESSAGE_COUNT", "Calibration verification numbers (CVN) message count for PID 06. Only for ISO 9141-2, ISO 14230-4 and SAE J1850.", None, None, None)
    CVN = C(0x06, 0x00, "CVN", "Calibration Verification Numbers (CVN) Several CVN can be output (4 bytes each) the number of CVN and CALID must match", None, None, None)
    IN_USE_PERF_TRACKING_MESSAGE_COUNT = C(0x07, 0x01, "IN_USE_PERF_TRACKING_MESSAGE_COUNT", "In-use performance tracking message count for PID 08 and 0A. Only for ISO 9141-2, ISO 14230-4 and SAE J1850.", 8, 10, None)
    IN_USE_PERF_TRACKING_SPARK_IGNITION = C(0x08, 0x04, "IN_USE_PERF_TRACKING_SPARK_IGNITION", "In-use performance tracking for spark ignition vehicles", None, None, None)
    ECU_NAME_MESSAGE_COUNT = C(0x09, 0x01, "ECU_NAME_MESSAGE_COUNT", "ECU name message count for PID 0A", None, None, None)
    ECU_NAME = C(0x0A, 0x14, "ECU_NAME", "ECU name", None, None, None)
    IN_USE_PERF_TRACKING_COMPRESSION_IGNITION = C(0x0B, 0x04, "IN_USE_PERF_TRACKING_COMPRESSION_IGNITION", "In-use performance tracking for compression ignition vehicles", None, None, None)


from typing import Optional

from serial import Serial # type: ignore

from .transport_base import TransportBase

from ..basetypes import MISSING
from ..utils import override_class_attributes


class TransportPort(TransportBase):
    def __init__(self, **kwargs) -> None:
        self.overridable_attributes = {
            "port": MISSING,
            "baudrate": 38400,
            "timeout": 5.0,
            "write_timeout": 3.0,
        }

        self.port: str
        self.baudrate: int
        self.timeout: float
        self.write_timeout: float

        self.serial_conn: Optional[Serial] = None

        override_class_attributes(self, self.overridable_attributes, **kwargs)

        if self.port is MISSING:
            raise ValueError("Port must be specified for TransportPort.")


    def __repr__(self) -> str:
        return f"<TransportPort {self.port} at {self.baudrate} baud>"


    def is_connected(self) -> bool:
        return self.serial_conn is not None and self.serial_conn.is_open


    def connect(self, **kwargs) -> None:
        override_class_attributes(self, self.overridable_attributes, True, **kwargs)

        self.serial_conn = Serial(
            port=self.port,
            baudrate=self.baudrate,
            timeout=self.timeout,
            write_timeout=self.write_timeout,
            **kwargs,
        )

    def close(self) -> None:
        if self.serial_conn and self.serial_conn.is_open:
            self.serial_conn.close()
        self.serial_conn = None


    def write_bytes(self, query: bytes) -> None:
        if not self.serial_conn or not self.serial_conn.is_open:
            raise RuntimeError("Serial port is not connected.")
        self._clear_buffer()
        self.serial_conn.write(query)
        self.serial_conn.flush()

    def read_bytes(self, expected_seq: bytes = b'>', size: int = 4096) -> bytes:
        if not self.serial_conn or not self.serial_conn.is_open:
            raise RuntimeError("Serial port is not connected.")
        return self.serial_conn.read_until(expected=expected_seq, size=size)


    def _clear_buffer(self) -> None:
        if self.serial_conn:
            self.serial_conn.reset_input_buffer()


from enum import Enum


class Mode(Enum):
    NONE = ''
    """Special mode used for the REPEAT command"""

    AT = "AT"
    """Special mode to send AT commands"""

    REQUEST = 0x01
    """Request current data"""
    FREEZE_FRAME = 0x02
    """Request freeze frame data"""
    STATUS_DTC = 0x03
    """Request stored DTCs (Diagnostic Trouble Codes)"""
    CLEAR_DTC = 0x04
    """Clear/reset DTCs (Diagnostic Trouble Codes)"""
    O2_SENSOR = 0x05
    """Request oxygen sensor monitoring test results"""
    PENDING_DTC = 0x06
    """Request DTCs (Diagnostic Trouble Codes) pending"""
    CONTROL_MODULE = 0x07
    """Request control module information"""
    O2_SENSOR_TEST = 0x08
    """Request oxygen sensor test results"""
    VEHICLE_INFO = 0x09
    """Request vehicle information"""
    PERMANENT_DTC = 0x0A
    """Request permanent DTCs (Diagnostic Trouble Codes)"""